#pragma once

#include <functional>
#include <string>

#include <curl/curl.h>

#include "headers.hpp"

namespace coeurl {
struct Client;

struct Request {
    enum class Method {
        Delete,
        Get,
        Head,
        Options,
        Patch,
        Post,
        Put,
    };

    Request(Client *client, Method m, std::string url_);
    ~Request();
    Request(Request const &) = delete;
    void operator=(Request const &) = delete;
    Request(Request &&) = delete;
    void operator=(Request &&) = delete;

    Request &max_redirects(long amount);
    Request &verify_peer(bool verify);
    Request &request(std::string r, std::string contenttype = "application/octet-stream");
    Request &request_headers(const Headers &h);
    //! Timeout connection after the specified amount of seconds, if the server
    //! stops sending acks.
    Request &connection_timeout(long t);

    Request &on_complete(const std::function<void(const Request &)> handler);
    Request &on_upload_progress(std::function<void(size_t progress, size_t total)> handler);
    Request &on_download_progess(std::function<void(size_t progress, size_t total)> handler);

    std::string_view url() const { return url_; }
    std::string_view response() const { return response_; }
    int response_code() const;
    CURLcode error_code() const { return curl_error; }
    Headers response_headers() const { return response_headers_; }

  private:
    /* CURLOPT_WRITEFUNCTION */
    static size_t write_cb(void *ptr, size_t size, size_t nmemb, void *data);
    /* CURLOPT_READFUNCTION */
    static size_t read_cb(char *buffer, size_t size, size_t nitems, void *data);
    /* CURLOPT_HEADERFUNCTION */
    static size_t header_cb(char *buffer, size_t size, size_t nitems, void *data);
    /* CURLOPT_PROGRESSFUNCTION */
    static int prog_cb(void *p, curl_off_t dltotal, curl_off_t dlnow, curl_off_t ult, curl_off_t uln);

    CURL *easy;
    std::string request_;
    std::string request_contenttype_;
    std::string response_;
    std::string url_;
    Headers response_headers_;
    curl_slist *request_headers_ = nullptr;

    Client *global;
    size_t readoffset = 0;
    char error[CURL_ERROR_SIZE];

    enum class Status {
        Running,
        Canceled,
        Done,
    } status = Status::Running;
    CURLcode curl_error = CURLcode::CURLE_OK;
    Method method = Method::Get;

    long connection_timeout_ = 0;

    std::function<void(const Request &)> on_complete_;
    std::function<void(size_t progress, size_t total)> on_upload_progress_, on_download_progess_;

    friend Client;
};
} // namespace coeurl
